/*
 * Copyright (c) 2005 Mellanox Technologies.  All rights reserved.
 * Copyright (c) 2005 SilverStorm Technologies.  All rights reserved.
 *
 * This software is available to you under the OpenIB.org BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: ipoib_ibat.c 195 2005-12-06 21:41:40Z ftillier $
 */


#include "ipoib_driver.h"
#include "ipoib_adapter.h"
#include "ipoib_port.h"
#include "ipoib_debug.h"
#include <iba/ib_at_ioctl.h>


static NTSTATUS
__ipoib_create(
	IN				DEVICE_OBJECT* const		pDevObj,
	IN				IRP* const					pIrp );

static NTSTATUS
__ipoib_cleanup(
	IN				DEVICE_OBJECT* const		pDevObj,
	IN				IRP* const					pIrp );

static NTSTATUS
__ipoib_close(
	IN				DEVICE_OBJECT* const		pDevObj,
	IN				IRP* const					pIrp );

static NTSTATUS
__ipoib_dispatch(
	IN				DEVICE_OBJECT* const		pDevObj,
	IN				IRP* const					pIrp );


static NTSTATUS
__ibat_get_ports(
	IN				IRP							*pIrp,
	IN				IO_STACK_LOCATION			*pIoStack )
{
	IOCTL_IBAT_PORTS_IN		*pIn;
	IOCTL_IBAT_PORTS_OUT	*pOut;
	KLOCK_QUEUE_HANDLE		hdl;
	cl_list_item_t			*pItem;
	ipoib_adapter_t			*pAdapter;
	LONG					nPorts;

	IPOIB_ENTER(IPOIB_DBG_IOCTL);

	if( pIoStack->Parameters.DeviceIoControl.InputBufferLength !=
		sizeof(IOCTL_IBAT_PORTS_IN) )
	{
		IPOIB_TRACE_EXIT( IPOIB_DBG_ERROR, ("Invalid input buffer size.\n") );
		return STATUS_INVALID_PARAMETER;
	}
	
	if( pIoStack->Parameters.DeviceIoControl.OutputBufferLength <
		sizeof(IOCTL_IBAT_PORTS_OUT) )
	{
		IPOIB_TRACE_EXIT( IPOIB_DBG_ERROR, ("Invalid output buffer size.\n") );
		return STATUS_INVALID_PARAMETER;
	}

	pIn = pIrp->AssociatedIrp.SystemBuffer;
	pOut = pIrp->AssociatedIrp.SystemBuffer;

	if( pIn->Version != IBAT_IOCTL_VERSION )
	{
		IPOIB_TRACE_EXIT( IPOIB_DBG_ERROR, ("Invalid version.\n") );
		return STATUS_INVALID_PARAMETER;
	}

	KeAcquireInStackQueuedSpinLock( &g_ipoib.lock, &hdl );
	nPorts = (LONG)cl_qlist_count( &g_ipoib.adapter_list );
	switch( nPorts )
	{
	case 0:
		cl_memclr( pOut->Ports, sizeof(pOut->Ports) );
		/* Fall through */
	case 1:
		pOut->Size = sizeof(IOCTL_IBAT_PORTS_OUT);
		break;

	default:
		pOut->Size = sizeof(IOCTL_IBAT_PORTS_OUT) + 
			(sizeof(IBAT_PORT_RECORD) * (nPorts - 1));
		break;
	}

	pIrp->IoStatus.Information = pOut->Size;

	if( pOut->Size > pIoStack->Parameters.DeviceIoControl.OutputBufferLength )
	{
		nPorts = 1 +
			(pIoStack->Parameters.DeviceIoControl.OutputBufferLength -
			sizeof(IOCTL_IBAT_PORTS_OUT)) / sizeof(IBAT_PORT_RECORD);

		pIrp->IoStatus.Information = sizeof(IOCTL_IBAT_PORTS_OUT) +
			((nPorts - 1) * sizeof(IBAT_PORT_RECORD));
	}

	pOut->NumPorts = 0;
	pItem = cl_qlist_head( &g_ipoib.adapter_list );
	while( pOut->NumPorts != nPorts )
	{
		pAdapter = CONTAINING_RECORD( pItem, ipoib_adapter_t, entry );
		pOut->Ports[pOut->NumPorts].CaGuid = pAdapter->guids.ca_guid;
		pOut->Ports[pOut->NumPorts].PortGuid = pAdapter->guids.port_guid;
		pOut->NumPorts++;

		pItem = cl_qlist_next( pItem );
	}

	KeReleaseInStackQueuedSpinLock( &hdl );
	return STATUS_SUCCESS;
}


static NTSTATUS
__ibat_get_ips(
	IN				IRP							*pIrp,
	IN				IO_STACK_LOCATION			*pIoStack )
{
	IOCTL_IBAT_IP_ADDRESSES_IN	*pIn;
	IOCTL_IBAT_IP_ADDRESSES_OUT	*pOut;
	KLOCK_QUEUE_HANDLE			hdl;
	cl_list_item_t				*pItem;
	ipoib_adapter_t				*pAdapter;
	LONG						nIps, maxIps;
	size_t						idx;
	net_address_item_t			*pAddr;
	UINT64						PortGuid;

	IPOIB_ENTER(IPOIB_DBG_IOCTL);

	if( pIoStack->Parameters.DeviceIoControl.InputBufferLength !=
		sizeof(IOCTL_IBAT_IP_ADDRESSES_IN) )
	{
		IPOIB_TRACE_EXIT( IPOIB_DBG_ERROR, ("Invalid input buffer size.\n") );
		return STATUS_INVALID_PARAMETER;
	}
	
	if( pIoStack->Parameters.DeviceIoControl.OutputBufferLength <
		sizeof(IOCTL_IBAT_IP_ADDRESSES_OUT) )
	{
		IPOIB_TRACE_EXIT( IPOIB_DBG_ERROR, ("Invalid output buffer size.\n") );
		return STATUS_INVALID_PARAMETER;
	}

	pIn = pIrp->AssociatedIrp.SystemBuffer;
	pOut = pIrp->AssociatedIrp.SystemBuffer;

	if( pIn->Version != IBAT_IOCTL_VERSION )
	{
		IPOIB_TRACE_EXIT( IPOIB_DBG_ERROR, ("Invalid version.\n") );
		return STATUS_INVALID_PARAMETER;
	}

	PortGuid = pIn->PortGuid;

	nIps = 0;
	pOut->AddressCount = 0;
	maxIps = 1 +
		((pIoStack->Parameters.DeviceIoControl.OutputBufferLength -
		sizeof(IOCTL_IBAT_IP_ADDRESSES_OUT)) / sizeof(IP_ADDRESS));

	KeAcquireInStackQueuedSpinLock( &g_ipoib.lock, &hdl );
	for( pItem = cl_qlist_head( &g_ipoib.adapter_list );
		pItem != cl_qlist_end( &g_ipoib.adapter_list );
		pItem = cl_qlist_next( pItem ) )
	{
		pAdapter = CONTAINING_RECORD( pItem, ipoib_adapter_t, entry );
		if( PortGuid && pAdapter->guids.port_guid != PortGuid )
			continue;

		cl_obj_lock( &pAdapter->obj );
		nIps += (LONG)cl_vector_get_size( &pAdapter->ip_vector );

		for( idx = 0;
			idx < cl_vector_get_size( &pAdapter->ip_vector );
			idx++ )
		{
			if( pOut->AddressCount == maxIps )
				break;

			pAddr = (net_address_item_t*)
				cl_vector_get_ptr( &pAdapter->ip_vector, idx );

			pOut->Address[pOut->AddressCount].IpVersion = 4;
			cl_memclr( &pOut->Address[pOut->AddressCount].Address,
				sizeof(IP_ADDRESS) );
			cl_memcpy( &pOut->Address[pOut->AddressCount].Address[12],
				pAddr->address.as_bytes, IPV4_ADDR_SIZE );

			pOut->AddressCount++;
		}
		cl_obj_unlock( &pAdapter->obj );
	}

	pOut->Size = sizeof(IOCTL_IBAT_IP_ADDRESSES_OUT);
	if( --nIps )
		pOut->Size += sizeof(IP_ADDRESS) * nIps;

	pIrp->IoStatus.Information = sizeof(IOCTL_IBAT_IP_ADDRESSES_OUT);
	if( --maxIps < nIps )
		pIrp->IoStatus.Information += (sizeof(IP_ADDRESS) * maxIps);
	else
		pIrp->IoStatus.Information += (sizeof(IP_ADDRESS) * nIps);

	KeReleaseInStackQueuedSpinLock( &hdl );
	return STATUS_SUCCESS;
}


static NTSTATUS
__ibat_mac_to_gid(
	IN				IRP							*pIrp,
	IN				IO_STACK_LOCATION			*pIoStack )
{
	NTSTATUS					status = STATUS_INVALID_PARAMETER;
	IOCTL_IBAT_MAC_TO_GID_IN	*pIn;
	IOCTL_IBAT_MAC_TO_GID_OUT	*pOut;
	KLOCK_QUEUE_HANDLE			hdl;
	cl_list_item_t				*pItem;
	ipoib_adapter_t				*pAdapter;

	IPOIB_ENTER(IPOIB_DBG_IOCTL);

	if( pIoStack->Parameters.DeviceIoControl.InputBufferLength !=
		sizeof(IOCTL_IBAT_MAC_TO_GID_IN) )
	{
		IPOIB_TRACE_EXIT( IPOIB_DBG_ERROR, ("Invalid input buffer size.\n") );
		return STATUS_INVALID_PARAMETER;
	}
	
	if( pIoStack->Parameters.DeviceIoControl.OutputBufferLength !=
		sizeof(IOCTL_IBAT_MAC_TO_GID_OUT) )
	{
		IPOIB_TRACE_EXIT( IPOIB_DBG_ERROR, ("Invalid output buffer size.\n") );
		return STATUS_INVALID_PARAMETER;
	}

	pIn = pIrp->AssociatedIrp.SystemBuffer;
	pOut = pIrp->AssociatedIrp.SystemBuffer;

	if( pIn->Version != IBAT_IOCTL_VERSION )
	{
		IPOIB_TRACE_EXIT( IPOIB_DBG_ERROR, ("Invalid version.\n") );
		return STATUS_INVALID_PARAMETER;
	}

	KeAcquireInStackQueuedSpinLock( &g_ipoib.lock, &hdl );

	for( pItem = cl_qlist_head( &g_ipoib.adapter_list );
		pItem != cl_qlist_end( &g_ipoib.adapter_list );
		pItem = cl_qlist_next( pItem ) )
	{
		pAdapter = CONTAINING_RECORD( pItem, ipoib_adapter_t, entry );
		if( pIn->PortGuid != pAdapter->guids.port_guid )
			continue;

		/* Found the port - lookup the MAC. */
		cl_obj_lock( &pAdapter->obj );
		if( pAdapter->p_port )
		{
			status = ipoib_mac_to_gid(
				pAdapter->p_port, *(mac_addr_t*)pIn->DestMac, &pOut->DestGid );
			if( NT_SUCCESS( status ) )
			{
				pIrp->IoStatus.Information =
					sizeof(IOCTL_IBAT_MAC_TO_GID_OUT);
			}
		}
		cl_obj_unlock( &pAdapter->obj );
		break;
	}

	KeReleaseInStackQueuedSpinLock( &hdl );

	return status;
}


void
ipoib_ref_ibat()
{
	NDIS_STATUS			status;
	NDIS_STRING			DeviceName;
	NDIS_STRING			DeviceLinkUnicodeString;
	PDRIVER_DISPATCH	DispatchTable[IRP_MJ_MAXIMUM_FUNCTION+1];
	DEVICE_OBJECT		*p_dev_obj;

	IPOIB_ENTER( IPOIB_DBG_IOCTL );

	if( InterlockedIncrement( &g_ipoib.ibat_ref ) == 1 )
	{
		NdisInitUnicodeString( &DeviceName, IBAT_DEV_NAME );
		NdisInitUnicodeString( &DeviceLinkUnicodeString, IBAT_DOS_DEV_NAME );

		NdisZeroMemory( DispatchTable, sizeof(DispatchTable) );

		DispatchTable[IRP_MJ_CREATE] = __ipoib_create;
		DispatchTable[IRP_MJ_CLEANUP] = __ipoib_cleanup;
		DispatchTable[IRP_MJ_CLOSE] = __ipoib_close;
		DispatchTable[IRP_MJ_DEVICE_CONTROL] = __ipoib_dispatch;
		DispatchTable[IRP_MJ_INTERNAL_DEVICE_CONTROL] = __ipoib_dispatch;

		status = NdisMRegisterDevice( g_ipoib.h_ndis_wrapper,
			&DeviceName, &DeviceLinkUnicodeString, &DispatchTable[0],
			&p_dev_obj, &g_ipoib.h_ibat_dev );
		if( status != NDIS_STATUS_SUCCESS )
		{
			IPOIB_TRACE( IPOIB_DBG_ERROR, 
				("NdisMRegisterDevice failed with status of %d\n", status) );
		}
	}

	IPOIB_EXIT( IPOIB_DBG_IOCTL );
}


void
ipoib_deref_ibat()
{
	IPOIB_ENTER( IPOIB_DBG_IOCTL );

	if( InterlockedDecrement( &g_ipoib.ibat_ref ) )
	{
		IPOIB_EXIT( IPOIB_DBG_IOCTL );
		return;
	}

	if( g_ipoib.h_ibat_dev )
	{
		NdisMDeregisterDevice( g_ipoib.h_ibat_dev );
		g_ipoib.h_ibat_dev = NULL;
	}

	IPOIB_EXIT( IPOIB_DBG_IOCTL );
}


static NTSTATUS
__ipoib_create(
	IN				DEVICE_OBJECT* const		pDevObj,
	IN				IRP* const					pIrp )
{
	IPOIB_ENTER( IPOIB_DBG_IOCTL );

	UNREFERENCED_PARAMETER( pDevObj );

	ipoib_ref_ibat();

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	IPOIB_EXIT( IPOIB_DBG_IOCTL );
	return STATUS_SUCCESS;
}


static NTSTATUS
__ipoib_cleanup(
	IN				DEVICE_OBJECT* const		pDevObj,
	IN				IRP* const					pIrp )
{
	IPOIB_ENTER( IPOIB_DBG_IOCTL );

	UNREFERENCED_PARAMETER( pDevObj );

	ipoib_deref_ibat();

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	IPOIB_EXIT( IPOIB_DBG_IOCTL );
	return STATUS_SUCCESS;
}


static NTSTATUS
__ipoib_close(
	IN				DEVICE_OBJECT* const		pDevObj,
	IN				IRP* const					pIrp )
{
	IPOIB_ENTER( IPOIB_DBG_IOCTL );

	UNREFERENCED_PARAMETER( pDevObj );

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	IPOIB_EXIT( IPOIB_DBG_IOCTL );
	return STATUS_SUCCESS;
}


static NTSTATUS
__ipoib_dispatch(
	IN				DEVICE_OBJECT* const		pDevObj,
	IN				IRP* const					pIrp )
{
	IO_STACK_LOCATION	*pIoStack;
	NTSTATUS			status = STATUS_SUCCESS;

	IPOIB_ENTER( IPOIB_DBG_IOCTL );

	UNREFERENCED_PARAMETER( pDevObj );

	pIoStack = IoGetCurrentIrpStackLocation( pIrp );

	pIrp->IoStatus.Information = 0;

	switch( pIoStack->Parameters.DeviceIoControl.IoControlCode )
	{
	case IOCTL_IBAT_PORTS:
		IPOIB_TRACE( IPOIB_DBG_IOCTL, ("IOCTL_IBAT_PORTS recieved\n") );
		status = __ibat_get_ports( pIrp, pIoStack );
		break;

	case IOCTL_IBAT_IP_ADDRESSES:
		IPOIB_TRACE( IPOIB_DBG_IOCTL, ("IOCTL_IBAT_IP_ADDRESSES recieved\n" ));
		status = __ibat_get_ips( pIrp, pIoStack );
		break;

	case IOCTL_IBAT_MAC_TO_GID:
		IPOIB_TRACE( IPOIB_DBG_IOCTL, ("IOCTL_IBAT_MAC_TO_GID recieved\n" ));
		status = __ibat_mac_to_gid( pIrp, pIoStack );
		break;

	default:
		IPOIB_TRACE( IPOIB_DBG_WARN, ("unknow IOCTL code = 0x%x\n",
			pIoStack->Parameters.DeviceIoControl.IoControlCode) );
		status = STATUS_INVALID_PARAMETER;
	}

	pIrp->IoStatus.Status = status;
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	IPOIB_EXIT( IPOIB_DBG_IOCTL );
	return status;
}

