/*
 * Copyright (c) 2005 SilverStorm Technologies.  All rights reserved.
 *
 * This software is available to you under the OpenIB.org BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: ipoib_xfr_mgr.h 800 2005-12-06 06:58:59Z sleybo $
 */


#ifndef _IPOIB_XFR_MGR_H_
#define _IPOIB_XFR_MGR_H_


#include <iba/ib_al.h>
#include <complib/cl_types.h>
#include <complib/cl_qpool.h>
#include <complib/cl_spinlock.h>
#include <complib/cl_qlist.h>
#include <complib/cl_qpool.h>
#include <complib/cl_list.h>


#include "ipoib_driver.h"
#include "ip_stats.h"
#include "ip_packet.h"


#include <complib/cl_packon.h>
/****s* IPoIB Driver/ipoib_hw_addr_t
* NAME
*   ipoib_hw_addr_t
*
* DESCRIPTION
*   The ipoib_hw_addr_t structure defines an IPoIB compatible hardware
*   address.  Values in this structure are stored in network order.
*
* SYNOPSIS
*/
typedef struct _ipoib_hw_addr
{
	uint32_t	flags_qpn;
	ib_gid_t	gid;

}	PACK_SUFFIX ipoib_hw_addr_t;
/*
* FIELDS
*	flags_qpn
*		Flags and queue pair number.  Use ipoib_addr_get_flags,
*		ipoib_addr_set_flags, ipoib_addr_set_qpn, and ipoib_addr_get_qpn
*		to manipulate the contents.
*
*	gid
*		IB GID value.
*
* SEE ALSO
*	IPoIB, ipoib_addr_get_flags, ipoib_addr_set_flags, ipoib_addr_set_qpn,
*	ipoib_addr_get_qpn
*********/
#include <complib/cl_packoff.h>



#ifdef __cplusplus
extern "C"
{
#endif


/*
 * Address accessors
 */

static inline uint8_t
ipoib_addr_get_flags(
	IN		const	ipoib_hw_addr_t* const		p_addr )
{
	return (uint8_t)(cl_ntoh32( p_addr->flags_qpn ) >> 24);
}

static inline void
ipoib_addr_set_flags(
	IN				ipoib_hw_addr_t* const		p_addr,
	IN		const	uint8_t						flags )
{
	p_addr->flags_qpn &= cl_ntoh32( 0xFFFFFF00 );
	p_addr->flags_qpn |= cl_ntoh32( flags );
}

static inline net32_t
ipoib_addr_get_qpn(
	IN		const	ipoib_hw_addr_t* const		p_addr )
{
	return cl_ntoh32( cl_ntoh32( p_addr->flags_qpn ) >> 8 );
}

static inline void
ipoib_addr_set_qpn(
	IN				ipoib_hw_addr_t* const		p_addr,
	IN		const	net32_t						qpn )
{
	p_addr->flags_qpn = cl_ntoh32( (cl_ntoh32(
		p_addr->flags_qpn ) & 0x000000FF ) | (cl_ntoh32( qpn ) << 8) );
}


/****f* IPOIB/ipoib_mac_from_sst_guid
* NAME
*	ipoib_mac_from_sst_guid
*
* DESCRIPTION
*	Generates an ethernet MAC address given a SilverStorm port GUID.
*
* SYNOPSIS
*/
static inline ib_api_status_t
ipoib_mac_from_sst_guid(
	IN		const	net64_t						port_guid,
		OUT			mac_addr_t* const			p_mac_addr )
{
	const uint8_t	*p_guid = (const uint8_t*)&port_guid;
	uint32_t		low24;

	/* Port guid is in network byte order.  OUI is in lower 3 bytes. */
	ASSERT( p_guid[0] == 0x00 && p_guid[1] == 0x06 && p_guid[2] == 0x6a );

	/*
	 * We end up using only the lower 23-bits of the GUID.  Trap that
	 * the 24th (bit 23) through 27th (bit 26) bit aren't set.
	 */
	if( port_guid & CL_HTON64( 0x0000000007800000 ) )
		return IB_INVALID_GUID;

	low24 = 0x00FFF000 -
		((((uint32_t)cl_ntoh64( port_guid ) & 0x00FFFFFF) - 0x101) * 2);
	low24 -= p_guid[3]; /* minus port number */

	p_mac_addr->addr[0] = p_guid[0];
	p_mac_addr->addr[1] = p_guid[1];
	p_mac_addr->addr[2] = p_guid[2];
	p_mac_addr->addr[3] = (uint8_t)(low24 >> 16);
	p_mac_addr->addr[4] = (uint8_t)(low24 >> 8);
	p_mac_addr->addr[5] = (uint8_t)low24;
	
	return IB_SUCCESS;
}
/*
* PARAMETERS
*	port_guid
*		The port GUID, in network byte order, for which to generate a
*		MAC address.
*
*	p_mac_addr
*		Pointer to a mac address in which to store the results.
*
* RETURN VALUES
*	IB_SUCCESS
*		The MAC address was successfully converted.
*
*	IB_INVALID_GUID
*		The port GUID provided was not a known GUID format.
*
* NOTES
*	The algorithm to convert portGuid to MAC address is as per DN0074, and
*	assumes a 2 port HCA.
*
* SEE ALSO
*	IPOIB
*********/


/****f* IPOIB/ipoib_mac_from_mlx_guid
* NAME
*	ipoib_mac_from_sst_guid
*
* DESCRIPTION
*	Generates an ethernet MAC address given a Mellanox port GUID.
*
* SYNOPSIS
*/
static inline ib_api_status_t
ipoib_mac_from_mlx_guid(
	IN		const	net64_t						port_guid,
		OUT			mac_addr_t* const			p_mac_addr )
{
	const uint8_t	*p_guid = (const uint8_t*)&port_guid;
	uint32_t		low24;

	/* Port guid is in network byte order.  OUI is in lower 3 bytes. */
	ASSERT( p_guid[0] == 0x00 && p_guid[1] == 0x02 && p_guid[2] == 0xc9 );

	if( (port_guid & CL_HTON64( 0x000000ffff0000 )) !=
		CL_HTON64(0x00000002000000))
	{
		return IB_INVALID_GUID;
	}

	low24 = ((uint32_t)cl_ntoh64( port_guid ) & 0x00FFFFFF);

	p_mac_addr->addr[0] = p_guid[0];
	p_mac_addr->addr[1] = p_guid[1];
	p_mac_addr->addr[2] = p_guid[2];
	p_mac_addr->addr[3] = (uint8_t)(low24 >> 16);
	p_mac_addr->addr[4] = (uint8_t)(low24 >> 8);
	p_mac_addr->addr[5] = (uint8_t)low24;
	
	return IB_SUCCESS;
}
/*
* PARAMETERS
*	port_guid
*		The port GUID, in network byte order, for which to generate a
*		MAC address.
*
*	p_mac_addr
*		Pointer to a mac address in which to store the results.
*
* RETURN VALUES
*	IB_SUCCESS
*		The MAC address was successfully converted.
*
*	IB_INVALID_GUID
*		The port GUID provided was not a known GUID format.
*
* NOTES
*	The algorithm to convert portGuid to MAC address is as per DN0074, and
*	assumes a 2 port HCA.
*
* SEE ALSO
*	IPOIB
*********/


/****f* IPOIB/ipoib_mac_from_guid
* NAME
*	ipoib_mac_from_guid
*
* DESCRIPTION
*	Generates an ethernet MAC address given a port GUID.
*
* SYNOPSIS
*/
static inline ib_api_status_t
ipoib_mac_from_guid(
	IN		const	net64_t						port_guid,
		OUT			mac_addr_t* const			p_mac_addr )
{
	ib_api_status_t	status;
	const uint8_t	*p_guid = (const uint8_t*)&port_guid;
	uint32_t		laa;

	/* Port guid is in network byte order.  OUI is in lower 3 bytes. */
	if( p_guid[0] == 0x00 && p_guid[1] == 0x06 && p_guid[2] == 0x6a )
	{
		status = ipoib_mac_from_sst_guid( port_guid, p_mac_addr );
		if( status == IB_SUCCESS )
			return IB_SUCCESS;
	}
	if( p_guid[0] == 0x00 && p_guid[1] == 0x02 && p_guid[2] == 0xc9 )
	{
		status = ipoib_mac_from_mlx_guid( port_guid, p_mac_addr );
		if( status == IB_SUCCESS )
			return IB_SUCCESS;
	}
	
	/* Value of zero is reserved. */
	laa = cl_atomic_inc( &g_ipoib.laa_idx );

	if( !laa )
		return IB_INVALID_GUID;

	p_mac_addr->addr[0] = 2; /* LAA bit */
	p_mac_addr->addr[1] = 0;
	p_mac_addr->addr[2] = (uint8_t)(laa >> 24);
	p_mac_addr->addr[3] = (uint8_t)(laa >> 16);
	p_mac_addr->addr[4] = (uint8_t)(laa >> 8);
	p_mac_addr->addr[5] = (uint8_t)laa;
	
	return IB_SUCCESS;
}
/*
* PARAMETERS
*	port_guid
*		The port GUID, in network byte order, for which to generate a
*		MAC address.
*
*	p_mac_addr
*		Pointer to a mac address in which to store the results.
*
* RETURN VALUES
*	IB_SUCCESS
*		The MAC address was successfully converted.
*
*	IB_INVALID_GUID
*		The port GUID provided was not a known GUID format.
*
* NOTES
*	Creates a locally administered address using a global incrementing counter.
*
* SEE ALSO
*	IPOIB
*********/


/****f* IPOIB/ipoib_sst_guid_from_mac
* NAME
*	ipoib_sst_guid_from_mac
*
* DESCRIPTION
*	Generates a port GUID given an ethernet MAC address.
*
* SYNOPSIS
*/
static inline ib_api_status_t
ipoib_sst_guid_from_mac(
	IN		const	mac_addr_t					mac,
		OUT			net64_t* const				p_port_guid )
{
	uint8_t		*p_guid = (uint8_t*)p_port_guid;
	uint32_t	low24;

	/* MAC address is in network byte order.  OUI is in lower 3 bytes. */
	if( mac.addr[0] != 0x00 || 
		mac.addr[1] != 0x06 || 
		mac.addr[2] != 0x6a )
	{
		return IB_INVALID_GUID;
	}

	low24 = mac.addr[3] << 16 || mac.addr[4] << 8 || mac.addr[5];

	low24 = 0x00FFF000 - low24;
	/* Divide by two */
	low24 >>= 1;
	/* Add the serial number base offset. */
	low24 += 0x101;

	/* OUI */
	p_guid[0] = mac.addr[0];
	p_guid[1] = mac.addr[1];
	p_guid[2] = mac.addr[2];
	/* Port number */
	p_guid[3] = mac.addr[5] & 0x01;
	/* Type */
	p_guid[4] = 0x98;
	/* Serial Number */
	p_guid[5] = (uint8_t)(low24 >> 16);
	p_guid[6] = (uint8_t)(low24 >> 8);
	p_guid[7] = (uint8_t)low24;
	
	return IB_SUCCESS;
}
/*
* PARAMETERS
*	port_guid
*		The port GUID, in network byte order, for which to generate a
*		MAC address.
*
*	p_mac_addr
*		Pointer to a mac address in which to store the results.
*
* RETURN VALUES
*	IB_SUCCESS
*		The MAC address was successfully converted.
*
*	IB_INVALID_GUID
*		The port GUID provided was not a known GUID format.
*
* NOTES
*	The algorithm to convert portGuid to MAC address is as per DN0074, and
*	assumes a 2 port HCA.
*
* SEE ALSO
*	IPOIB
*********/


/****f* IPOIB/ipoib_mlx_guid_from_mac
* NAME
*	ipoib_mlx_guid_from_mac
*
* DESCRIPTION
*	Generates a port GUID given an ethernet MAC address.
*
* SYNOPSIS
*/
static inline ib_api_status_t
ipoib_mlx_guid_from_mac(
	IN		const	mac_addr_t					mac,
		OUT			net64_t* const				p_port_guid )
{
	uint8_t		*p_guid = (uint8_t*)p_port_guid;
	uint32_t	low24;

	/* MAC address is in network byte order.  OUI is in lower 3 bytes. */
	if( mac.addr[0] != 0x00 || 
		mac.addr[1] != 0x02 || 
		mac.addr[2] != 0xc9 )
	{
		return IB_INVALID_GUID;
	}

	low24 = mac.addr[3] << 16 || mac.addr[4] << 8 || mac.addr[5];

	/* OUI */
	p_guid[0] = mac.addr[0];
	p_guid[1] = mac.addr[1];
	p_guid[2] = mac.addr[2];
	p_guid[3] = 0x02;
	p_guid[4] = 0x00;
	/* Serial Number */
	p_guid[5] = (uint8_t)(low24 >> 16);
	p_guid[6] = (uint8_t)(low24 >> 8);
	p_guid[7] = (uint8_t)low24;
	
	return IB_SUCCESS;
}
/*
* PARAMETERS
*	port_guid
*		The port GUID, in network byte order, for which to generate a
*		MAC address.
*
*	p_mac_addr
*		Pointer to a mac address in which to store the results.
*
* RETURN VALUES
*	IB_SUCCESS
*		The MAC address was successfully converted.
*
*	IB_INVALID_GUID
*		The port GUID provided was not a known GUID format.
*
* NOTES
*	The algorithm to convert portGuid to MAC address is as 
*
* SEE ALSO
*	IPOIB
*********/


#ifdef __cplusplus
}
#endif

#endif	/* _IPOIB_XFR_MGR_H_ */
